@@ -21,13 +21,14 @@ import time
 from functools import wraps
 from io import BytesIO
 
-import jsonpath_rw
 import requests
-from lxml import etree
+from jsonpath_ng.ext import parse as jsonpath_parse
+from lxml import etree, html
+from requests.structures import CaseInsensitiveDict
 
 import apiritif
-from apiritif.utilities import *
 from apiritif.thread import get_from_thread_store, put_into_thread_store
+from apiritif.utilities import *
 from apiritif.utils import headers_as_text, assert_regexp, assert_not_regexp, log, get_trace
 
 
@@ -48,7 +49,8 @@ class http(object):
 
     @staticmethod
     def request(method, address, session=None,
-                params=None, headers=None, cookies=None, data=None, json=None, allow_redirects=True, timeout=30):
+                params=None, headers=None, cookies=None, data=None, json=None, files=None,
+                allow_redirects=True, timeout=30):
         """
 
         :param method: str
@@ -57,8 +59,8 @@ class http(object):
         :rtype: HTTPResponse
         """
         http.log.info("Request: %s %s", method, address)
-        msg = "Request: params=%r, headers=%r, cookies=%r, data=%r, json=%r, allow_redirects=%r, timeout=%r"
-        http.log.debug(msg, params, headers, cookies, data, json, allow_redirects, timeout)
+        msg = "Request: params=%r, headers=%r, cookies=%r, data=%r, json=%r, files=%r, allow_redirects=%r, timeout=%r"
+        http.log.debug(msg, params, headers, cookies, data, json, files, allow_redirects, timeout)
 
         if headers is None:
             headers = {}
@@ -68,16 +70,19 @@ class http(object):
         if session is None:
             session = requests.Session()
         request = requests.Request(method, address,
-                                   params=params, headers=headers, cookies=cookies, json=json, data=data)
+                                   params=params, headers=headers, cookies=cookies, json=json, data=data, files=files)
         prepared = session.prepare_request(request)
         settings = session.merge_environment_settings(prepared.url, {}, False, False, None)
         try:
             response = session.send(prepared, allow_redirects=allow_redirects, timeout=timeout, **settings)
-        except requests.exceptions.Timeout:
+        except requests.exceptions.Timeout as exc:
+            recorder.record_http_request_failure(method, address, prepared, exc, session)
             raise TimeoutError("Connection to %s timed out" % address)
-        except requests.exceptions.ConnectionError:
+        except requests.exceptions.ConnectionError as exc:
+            recorder.record_http_request_failure(method, address, prepared, exc, session)
             raise ConnectionError("Connection to %s failed" % address)
-        except BaseException:
+        except BaseException as exc:
+            recorder.record_http_request_failure(method, address, prepared, exc, session)
             raise
         http.log.info("Response: %s %s", response.status_code, response.reason)
         http.log.debug("Response headers: %r", response.headers)
@@ -220,7 +225,7 @@ class smart_transaction(transaction_logged):
         super(smart_transaction, self).__enter__()
         put_into_thread_store(test_case=self.name, test_suite=self.test_suite)
         for func in apiritif.get_transaction_handlers()["enter"]:
-            func(self.name, self.test_suite)    # params for compatibility, remove if bzt > 1.4.1 in cloud
+            func(self.name, self.test_suite)  # params for compatibility, remove if bzt > 1.4.1 in cloud
 
         self.controller.startTest()
 
@@ -254,34 +259,62 @@ class smart_transaction(transaction_logged):
 
 
 class Event(object):
-    def __init__(self):
+    def __init__(self, response=None):
         self.timestamp = time.time()
+        self.response = response
+
+    def to_dict(self):  # supposed to be overridden by extenders
+        return {}
 
 
 class Request(Event):
     def __init__(self, method, address, request, response, session):
         """
-
         :type method: str
         :type address: str
         :type request: requests.PreparedRequest
         :type response: HTTPResponse
         :type session: requests.Session
         """
-        super(Request, self).__init__()
+        super(Request, self).__init__(response)
         self.method = method
         self.address = address
         self.request = request
-        self.response = response
         self.session = session
 
     def __repr__(self):
         return "Request(method=%r, address=%r)" % (self.method, self.address)
 
 
+class RequestFailure(Request):
+    def __init__(self, method, address, request, exc, session):
+        """
+
+        :type method: str
+        :type address: str
+        :type request: requests.PreparedRequest
+        :type exc: BaseException
+        :type session: requests.Session
+        """
+        response = requests.Response()
+        response.request = request
+        response.status_code = 999
+        response._content = ""
+
+        super(RequestFailure, self).__init__(method, address, request, HTTPResponse(response), session)
+        self.method = method
+        self.address = address
+        self.request = request
+        self.exception = exc
+        self.session = session
+
+    def __repr__(self):
+        return "RequestFailure(method=%r, address=%r)" % (self.method, self.address)
+
+
 class TransactionStarted(Event):
     def __init__(self, transaction):
-        super(TransactionStarted, self).__init__()
+        super(TransactionStarted, self).__init__(None)
         self.transaction = transaction
         self.transaction_name = transaction.name
 
@@ -300,10 +333,10 @@ class TransactionEnded(Event):
 
 
 class Assertion(Event):
-    def __init__(self, name, response):
-        super(Assertion, self).__init__()
+    def __init__(self, name, response, extras):
+        super(Assertion, self).__init__(response)
         self.name = name
-        self.response = response
+        self.extras = extras
 
     def __repr__(self):
         return "Assertion(name=%r)" % self.name
@@ -311,9 +344,8 @@ class Assertion(Event):
 
 class AssertionFailure(Event):
     def __init__(self, assertion_name, response, failure_message):
-        super(AssertionFailure, self).__init__()
+        super(AssertionFailure, self).__init__(response)
         self.name = assertion_name
-        self.response = response
         self.failure_message = failure_message
 
     def __repr__(self):
@@ -364,8 +396,12 @@ class _EventRecorder(object):
     def record_http_request(self, method, address, request, response, session):
         self.record_event(Request(method, address, request, response, session))
 
-    def record_assertion(self, assertion_name, target_response):
-        self.record_event(Assertion(assertion_name, target_response))
+    def record_http_request_failure(self, method, address, request, exception, session):
+        failure = RequestFailure(method, address, request, exception, session)
+        self.record_event(failure)
+
+    def record_assertion(self, assertion_name, target_response, extras):
+        self.record_event(Assertion(assertion_name, target_response, extras))
 
     def record_assertion_failure(self, assertion_name, target_response, failure_message):
         self.record_event(AssertionFailure(assertion_name, target_response, failure_message))
@@ -375,7 +411,8 @@ class _EventRecorder(object):
         @wraps(assertion_method)
         def _impl(self, *method_args, **method_kwargs):
             assertion_name = getattr(assertion_method, '__name__', 'assertion')
-            recorder.record_assertion(assertion_name, self)
+            extras = {"args": list(method_args), "kwargs": method_kwargs}
+            recorder.record_assertion(assertion_name, self, extras)
             try:
                 return assertion_method(self, *method_args, **method_kwargs)
             except BaseException as exc:
@@ -447,7 +484,8 @@ class HTTPTarget(object):
         return addr
 
     def request(self, method, path,
-                params=None, headers=None, cookies=None, data=None, json=None, allow_redirects=None, timeout=None):
+                params=None, headers=None, cookies=None, data=None, json=None, files=None,
+                allow_redirects=None, timeout=None):
         """
         Prepares and sends an HTTP request. Returns the HTTPResponse object.
 
@@ -471,7 +509,7 @@ class HTTPTarget(object):
         req_headers.update(headers)
 
         response = http.request(method, address, session=self.__session,
-                                params=params, headers=req_headers, cookies=cookies, data=data, json=json,
+                                params=params, headers=req_headers, cookies=cookies, data=data, json=json, files=files,
                                 allow_redirects=allow_redirects, timeout=timeout)
         if self._auto_assert_ok:
             response.assert_ok()
@@ -515,8 +553,8 @@ class HTTPResponse(object):
         self.status_code = int(py_response.status_code)
         self.reason = py_response.reason
 
-        self.headers = dict(py_response.headers)
-        self.cookies = dict(py_response.cookies)
+        self.headers = CaseInsensitiveDict(py_response.headers)
+        self.cookies = {x: py_response.cookies.get(x) for x in py_response.cookies}
 
         self.text = py_response.text
         self.content = py_response.content
@@ -601,6 +639,15 @@ class HTTPResponse(object):
             raise AssertionError(msg)
         return self
 
+    @recorder.assertion_decorator
+    def assert_status_code_in(self, codes, msg=None):
+        actual = str(self.status_code)
+        expected = [str(code) for code in codes]
+        if actual not in expected:
+            msg = msg or "Actual status code (%s) is not one of expected expected (%s)" % (actual, expected)
+            raise AssertionError(msg)
+        return self
+
     @recorder.assertion_decorator
     def assert_not_status_code(self, code, msg=None):
         actual = str(self.status_code)
@@ -679,7 +726,7 @@ class HTTPResponse(object):
 
     @recorder.assertion_decorator
     def assert_jsonpath(self, jsonpath_query, expected_value=None, msg=None):
-        jsonpath_expr = jsonpath_rw.parse(jsonpath_query)
+        jsonpath_expr = jsonpath_parse(jsonpath_query)
         body = self.json()
         matches = jsonpath_expr.find(body)
         if not matches:
@@ -694,7 +741,7 @@ class HTTPResponse(object):
 
     @recorder.assertion_decorator
     def assert_not_jsonpath(self, jsonpath_query, msg=None):
-        jsonpath_expr = jsonpath_rw.parse(jsonpath_query)
+        jsonpath_expr = jsonpath_parse(jsonpath_query)
         body = self.json()
         matches = jsonpath_expr.find(body)
         if matches:
@@ -722,6 +769,28 @@ class HTTPResponse(object):
             raise AssertionError(msg)
         return self
 
+    @recorder.assertion_decorator
+    def assert_cssselect(self, query, expected_value=None, attribute=None, msg=None):
+        tree = html.fromstring(self.text)
+        q = tree.cssselect(query)
+        vals = [(x.text if attribute is None else x.attrib[attribute]) for x in q]
+
+        matches = expected_value in vals if expected_value is not None else vals
+        if not matches:
+            msg = msg or "CSSSelect query %r didn't match response content: %s" % (query, self.text)
+            raise AssertionError(msg)
+        return self
+
+    @recorder.assertion_decorator
+    def assert_not_cssselect(self, query, expected_value=None, attribute=None, msg=None):
+        try:
+            self.assert_cssselect(query, expected_value, attribute)
+        except AssertionError:
+            return self
+
+        msg = msg or "CSSSelect query %r did match response content: %s" % (query, self.text)
+        raise AssertionError(msg)
+
     # TODO: assertTiming? to assert response time / connection time
 
     def extract_regex(self, regex, default=None):
@@ -732,13 +801,22 @@ class HTTPResponse(object):
         return extracted_value
 
     def extract_jsonpath(self, jsonpath_query, default=None):
-        jsonpath_expr = jsonpath_rw.parse(jsonpath_query)
+        jsonpath_expr = jsonpath_parse(jsonpath_query)
         body = self.json()
         matches = jsonpath_expr.find(body)
         if not matches:
             return default
         return matches[0].value
 
+    def extract_cssselect(self, selector, attribute=None, default=None):
+        tree = html.fromstring(self.text)
+        q = tree.cssselect(selector)
+        matches = [(x.text if attribute is None else x.attrib[attribute]) for x in q]
+
+        if not matches:
+            return default
+        return matches[0]
+
     def extract_xpath(self, xpath_query, default=None, parser_type='html', validate=False):
         parser = etree.HTMLParser() if parser_type == 'html' else etree.XMLParser(dtd_validation=validate)
         tree = etree.parse(BytesIO(self.content), parser)
